from __future__ import annotations
from copy import deepcopy
from pathlib import Path
import re
from typing import Any, Optional
from .core import VCSDescription
from .errors import ConfigError, InvalidTagError
from .logging import log, warn_extra_fields
from .util import (
    bool_guard,
    optional_str_guard,
    split_pep440_version,
    split_version,
    str_guard,
    strip_prefix,
    strip_suffix,
)

#: The default formats for the ``"basic"`` ``format`` method
DEFAULT_FORMATS = {
    "distance": "{version}.post{distance}+{vcs}{rev}",
    "dirty": "{version}+d{build_date:%Y%m%d}",
    "distance-dirty": "{version}.post{distance}+{vcs}{rev}.d{build_date:%Y%m%d}",
}


def basic_tag2version(*, tag: str, params: dict[str, Any]) -> str:
    """Implements the ``"basic"`` ``tag2version`` method"""
    params = params.copy()
    try:
        rmprefix = str_guard(
            params.pop("rmprefix"), "tool.versioningit.tag2version.rmprefix"
        )
    except KeyError:
        pass
    else:
        tag = strip_prefix(tag, rmprefix)
    try:
        rmsuffix = str_guard(
            params.pop("rmsuffix"), "tool.versioningit.tag2version.rmsuffix"
        )
    except KeyError:
        pass
    else:
        tag = strip_suffix(tag, rmsuffix)
    require_match = bool_guard(
        params.pop("require-match", False),
        "tool.versioningit.tag2version.require-match",
    )
    try:
        regex = str_guard(params.pop("regex"), "tool.versioningit.tag2version.regex")
    except KeyError:
        pass
    else:
        m = re.search(regex, tag)
        if m is None:
            if require_match:
                raise InvalidTagError(f"tag2version.regex did not match tag {tag!r}")
            else:
                log.info(
                    "tag2version.regex did not match tag %r; leaving unmodified", tag
                )
        else:
            if "version" in m.groupdict():
                tag = m["version"]
            else:
                tag = m[0]
            if tag is None:
                raise InvalidTagError(
                    "'version' group in tool.versioningit.tag2version.regex did"
                    " not participate in match"
                )
    warn_extra_fields(
        params,
        "tool.versioningit.tag2version",
        ["rmprefix", "rmsuffix", "regex", "require-match"],
    )
    return tag.lstrip("v")


def basic_format(
    *,
    description: VCSDescription,
    base_version: str,
    next_version: str,
    params: dict[str, Any],
) -> str:
    """Implements the ``"basic"`` ``format`` method"""
    branch: Optional[str]
    if description.branch is not None:
        branch = re.sub(r"[^A-Za-z0-9.]", ".", description.branch)
    else:
        branch = None
    fields = {
        **description.fields,
        "branch": branch,
        "version": base_version,
        "base_version": base_version,
        "next_version": next_version,
    }
    formats = {**DEFAULT_FORMATS, **params}
    try:
        fmt = formats[description.state]
    except KeyError:
        raise ConfigError(
            f"No format string for {description.state!r} state found in"
            " tool.versioningit.format"
        )
    return fmt.format_map(fields)


def basic_write(
    *,
    project_dir: str | Path,
    template_fields: dict[str, Any],
    params: dict[str, Any],
) -> None:
    """Implements the ``"basic"`` ``write`` method"""
    params = params.copy()
    filename = str_guard(params.pop("file", None), "tool.versioningit.write.file")
    path = Path(project_dir, filename)
    encoding = str_guard(
        params.pop("encoding", "utf-8"), "tool.versioningit.write.encoding"
    )
    try:
        template = str_guard(params.pop("template"), "tool.versioningit.write.template")
    except KeyError:
        if path.suffix == ".py":
            template = '__version__ = "{version}"'
        elif path.suffix == ".txt" or path.suffix == "":
            template = "{version}"
        else:
            raise ConfigError(
                "tool.versioningit.write.template not specified and file has"
                f" unknown suffix {path.suffix!r}"
            )
    warn_extra_fields(
        params, "tool.versioningit.write", ["file", "encoding", "template"]
    )
    log.debug("Ensuring parent directories of %s exist", path)
    path.parent.mkdir(parents=True, exist_ok=True)
    log.info("Writing version to file %s", path)
    path.write_text(template.format_map(template_fields) + "\n", encoding=encoding)


def basic_template_fields(
    *,
    version: str,
    description: Optional[VCSDescription],
    base_version: Optional[str],
    next_version: Optional[str],
    params: dict[str, Any],
) -> dict[str, Any]:
    """Implements the ``"basic"`` ``template-fields`` method"""
    params = deepcopy(params)
    vtuple_params = params.pop("version-tuple", {})
    SUBTABLE = "tool.versioningit.template-fields.version-tuple"
    if not isinstance(vtuple_params, dict):
        raise ConfigError(f"{SUBTABLE} must be a table")
    pep440 = bool_guard(vtuple_params.pop("pep440", False), f"{SUBTABLE}.pep440")
    epoch: Optional[bool]
    try:
        epoch = bool_guard(vtuple_params.pop("epoch"), f"{SUBTABLE}.epoch")
    except KeyError:
        epoch = None
    else:
        if not pep440:
            log.warning("%s.epoch is ignored when pep440 is false", SUBTABLE)
    split_on = optional_str_guard(
        vtuple_params.pop("split-on", None), f"{SUBTABLE}.split-on"
    )
    if pep440 and split_on is not None:
        log.warning("%s.split-on is ignored when pep440 is true", SUBTABLE)
    double_quote = bool_guard(
        vtuple_params.pop("double-quote", True), f"{SUBTABLE}.double-quote"
    )
    warn_extra_fields(
        vtuple_params, SUBTABLE, ["pep440", "epoch", "split-on", "double-quote"]
    )
    warn_extra_fields(params, "tool.versioningit.template-fields", ["version-tuple"])
    if pep440:
        version_tuple = split_pep440_version(
            version, epoch=epoch, double_quote=double_quote
        )
    else:
        version_tuple = split_version(
            version, split_on=split_on, double_quote=double_quote
        )
    fields: dict[str, Any] = {}
    if description is not None:
        fields.update(description.fields)
        fields["branch"] = description.branch
    if base_version is not None:
        fields["base_version"] = base_version
    if next_version is not None:
        fields["next_version"] = next_version
    fields["version"] = version
    fields["version_tuple"] = version_tuple
    return fields



from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
import re
import subprocess
from typing import Any, NamedTuple, Optional
from .core import VCSDescription
from .errors import ConfigError, NoTagError, NotVCSError
from .logging import log, warn_extra_fields
from .util import (
    fromtimestamp,
    get_build_date,
    is_sdist,
    list_str_guard,
    optional_str_guard,
    readcmd,
    runcmd,
    str_guard,
)

DEFAULT_DATE = fromtimestamp(0)

# Values git-config accepts as true & false:
TRUTH_VALUES = {
    "yes": True,
    "on": True,
    "true": True,
    "1": True,
    "no": False,
    "off": False,
    "false": False,
    "0": False,
    "": False,
}


class Describe(NamedTuple):
    """
    Parsed representation of `git describe` output when it includes all three
    fields and does not end with "``-dirty``"
    """

    #: The most recent tag
    tag: str

    #: The number of commits since the tag
    distance: int

    #: The abbreviated hash of the HEAD commit
    rev: str

    @classmethod
    def parse(cls, s: str) -> Describe:
        m = re.fullmatch(r"(?P<tag>.+)-(?P<distance>[0-9]+)-g(?P<rev>[0-9a-f]+)?", s)
        if not m:
            raise ValueError("Could not parse `git describe` output")
        tag = m["tag"]
        assert isinstance(tag, str)
        distance = int(m["distance"])
        rev = m["rev"]
        assert isinstance(rev, str)
        return cls(tag, distance, rev)


@dataclass
class DescribeOpts:
    tags: bool
    match: list[str]
    exclude: list[str]

    @classmethod
    def parse_describe_subst(cls, s: str) -> DescribeOpts:
        m = re.fullmatch(r"\$Format:%\(describe(?::(?P<options>.*))?\)\$", s)
        if not m:
            raise ValueError(
                f"Expected string in format '$Format:%(describe[:options])$', got {s!r}"
            )
        tags = False
        match: list[str] = []
        exclude: list[str] = []
        options = m["options"]
        if options:
            # As of Git 2.35.1, though the docs say that %(describe) options
            # are comma-separated, they're actually comma-terminated, with
            # consecutive commas creating an empty option and empty options
            # causing the whole placeholder to be invalid.  Also, there's no
            # escaping support to worry about.
            if options.endswith(","):
                options = options[:-1]
            for opt in options.split(","):
                name, eq, value = opt.partition("=")
                if name == "tags":
                    if eq:
                        try:
                            tags = TRUTH_VALUES[value.lower()]
                        except KeyError:
                            # Git accepts invalid booleans and treats them as
                            # false, so we should, too, but we should at least
                            # warn the user that they probably made a mistake.
                            log.warning(
                                "Invalid boolean value for 'tags' option to"
                                " %%(describe) format: %r; treating as false",
                                value,
                            )
                            tags = False
                    else:
                        tags = True
                elif name == "match":
                    if not value:
                        raise ValueError(f"Option missing value: {opt!r}")
                    match.append(value)
                elif name == "exclude":
                    if not value:
                        raise ValueError(f"Option missing value: {opt!r}")
                    exclude.append(value)
                else:
                    raise ValueError(f"Unknown option: {opt!r}")
        return cls(tags=tags, match=match, exclude=exclude)

    def as_args(self) -> list[str]:
        args: list[str] = []
        if self.tags:
            args.append("--tags")
        for pat in self.match:
            args.append(f"--match={pat}")
        for pat in self.exclude:
            args.append(f"--exclude={pat}")
        return args


@dataclass
class GitRepo:
    """Methods for querying a Git repository"""

    #: The repository's working tree or a subdirectory thereof
    path: str | Path

    def ensure_is_repo(self) -> None:
        """
        Test whether `path` is under Git revision control; if it is not (or if
        Git is not installed), raise a `NotVCSError`
        """
        try:
            if (
                self.read(
                    "rev-parse", "--is-inside-work-tree", stderr=subprocess.DEVNULL
                )
                == "false"
            ):
                # We are inside a .git directory
                raise NotVCSError(f"{self.path} is not in a Git working tree")
        except FileNotFoundError:
            raise NotVCSError("Git not installed; assuming this isn't a Git repository")
        except subprocess.CalledProcessError:
            raise NotVCSError(f"{self.path} is not in a Git repository")
        try:
            # Check whether `path` is tracked by Git (Note that we can't rely
            # on this check alone, as it succeeds when inside a .git/
            # directory)
            runcmd(
                "git",
                "ls-files",
                "--error-unmatch",
                ".",
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                cwd=str(self.path),
            )
        except subprocess.CalledProcessError:
            raise NotVCSError(f"{self.path} is not tracked by Git")

    def read(self, *args: str, **kwargs: Any) -> str:
        """
        Run a Git command with the given arguments in `path` and return the
        stripped stdout
        """
        return readcmd("git", *args, cwd=str(self.path), **kwargs)

    def describe(self, opts: DescribeOpts) -> str:
        """
        Run ``git describe --long --dirty --always`` in the repository with the
        given options; if the command fails, raises `NoTagError`
        """
        try:
            return self.read(
                "describe",
                "--long",
                "--dirty",
                "--always",
                *opts.as_args(),
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            # As far as I'm aware, this only happens in a repo without any
            # commits or a corrupted repo.
            raise NoTagError(f"`git describe` command failed: {e.stderr.strip()}")

    def get_branch(self) -> Optional[str]:
        """
        Return the name of the current branch, or `None` if the repository is
        in a detached HEAD state
        """
        try:
            return self.read(
                "symbolic-ref", "--short", "-q", "HEAD", stderr=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError:
            return None


def describe_git(*, project_dir: str | Path, params: dict[str, Any]) -> VCSDescription:
    """Implements the ``"git"`` ``vcs`` method"""
    params = params.copy()
    match = list_str_guard(params.pop("match", []), "tool.versioningit.vcs.match")
    exclude = list_str_guard(params.pop("exclude", []), "tool.versioningit.vcs.exclude")
    default_tag = optional_str_guard(
        params.pop("default-tag", None), "tool.versioningit.vcs.default-tag"
    )
    warn_extra_fields(
        params, "tool.versioningit.vcs", ["match", "exclude", "default-tag"]
    )
    build_date = get_build_date()
    repo = GitRepo(project_dir)
    repo.ensure_is_repo()
    vdesc = describe_git_core(
        repo,
        build_date,
        default_tag,
        DescribeOpts(tags=True, match=match, exclude=exclude),
    )
    if "revision" not in vdesc.fields:
        revision, author_ts, committer_ts = repo.read(
            "--no-pager", "show", "-s", "--format=%H%n%at%n%ct"
        ).splitlines()
        vdesc.fields["revision"] = revision
        vdesc.fields["author_date"] = fromtimestamp(int(author_ts))
        vdesc.fields["committer_date"] = fromtimestamp(int(committer_ts))
    return vdesc


def describe_git_archive(
    *, project_dir: str | Path, params: dict[str, Any]
) -> VCSDescription:
    """Implements the ``"git-archive"`` ``vcs`` method"""
    params = params.copy()
    default_tag = optional_str_guard(
        params.pop("default-tag", None), "tool.versioningit.vcs.default-tag"
    )
    describe_subst = str_guard(
        params.pop("describe-subst", None), "tool.versioningit.vcs.describe-subst"
    )
    warn_extra_fields(
        params, "tool.versioningit.vcs", ["default-tag", "describe-subst"]
    )
    build_date = get_build_date()
    repo = GitRepo(project_dir)
    try:
        repo.ensure_is_repo()
    except NotVCSError:
        if is_sdist(project_dir):
            pass
        elif describe_subst == "":
            raise NoTagError(
                "tool.versioningit.vcs.describe-subst is empty in Git archive"
            )
        elif describe_subst.startswith("$Format"):
            raise NoTagError(
                "tool.versioningit.vcs.describe-subst not expanded in Git archive"
            )
        elif describe_subst.startswith("%(describe"):
            raise NoTagError(
                "tool.versioningit.vcs.describe-subst format was invalid,"
                f" expanded to {describe_subst!r}"
            )
        else:
            log.info(
                "Parsing version information from describe-subst = %r", describe_subst
            )
            try:
                tag, distance, rev = Describe.parse(describe_subst)
            except ValueError:
                tag = describe_subst
                distance = 0
                rev = "0" * 7
            return VCSDescription(
                tag=tag,
                state="distance" if distance else "exact",
                branch=None,
                fields={
                    "distance": distance,
                    "rev": rev,
                    "build_date": build_date,
                    "vcs": "g",
                    "vcs_name": "git",
                },
            )
        raise
    try:
        opts = DescribeOpts.parse_describe_subst(describe_subst)
    except ValueError as e:
        raise ConfigError(f"Invalid tool.versioningit.vcs.describe-subst value: {e}")
    vdesc = describe_git_core(repo, build_date, default_tag, opts)
    vdesc.fields.pop("revision", None)
    vdesc.fields.pop("author_date", None)
    vdesc.fields.pop("committer_date", None)
    return vdesc


def describe_git_core(
    repo: GitRepo,
    build_date: datetime,
    default_tag: Optional[str],
    opts: DescribeOpts,
) -> VCSDescription:
    """Common functionality of the ``"git"`` and ``"git-archive"`` methods"""
    try:
        description = repo.describe(opts)
    except NoTagError as e:
        # There are no commits in the repo
        if default_tag is not None:
            log.error("%s", e)
            log.info("Falling back to default tag %r", default_tag)
            return VCSDescription(
                tag=default_tag,
                state="dirty",
                branch=repo.get_branch(),
                fields={
                    "distance": 0,
                    "rev": "0" * 7,
                    "revision": "0" * 40,
                    "author_date": min(build_date, DEFAULT_DATE),
                    "committer_date": min(build_date, DEFAULT_DATE),
                    "build_date": build_date,
                    "vcs": "g",
                    "vcs_name": "git",
                },
            )
        else:
            raise
    if description.endswith("-dirty"):
        dirty = True
        description = description[: -len("-dirty")]
    else:
        dirty = False
    try:
        tag, distance, rev = Describe.parse(description)
    except ValueError:
        if default_tag is not None:
            log.info(
                "`git describe` returned a hash instead of a tag; falling back to"
                " default tag %r",
                default_tag,
            )
            tag = default_tag
            distance = int(repo.read("rev-list", "--count", "HEAD")) - 1
            rev = description
        else:
            raise NoTagError("`git describe` could not find a tag")
    if distance and dirty:
        state = "distance-dirty"
    elif distance:
        state = "distance"
    elif dirty:
        state = "dirty"
    else:
        state = "exact"
    return VCSDescription(
        tag=tag,
        state=state,
        branch=repo.get_branch(),
        fields={
            "distance": distance,
            "rev": rev,
            "build_date": build_date,
            "vcs": "g",
            "vcs_name": "git",
        },
    )



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING
from .core import Report, Versioningit
from .errors import NoTagError, NotSdistError, NotVersioningitError
from .logging import init_logging, log

if TYPE_CHECKING:
    from setuptools import Distribution


def setuptools_finalizer(dist: Distribution) -> None:
    """
    The entry point called by setuptools to retrieve the version for a project
    """
    init_logging()
    # PEP 517 says that "All hooks are run with working directory set to the
    # root of the source tree".
    PROJECT_ROOT = Path().resolve()
    log.info("Project dir: %s", PROJECT_ROOT)
    try:
        vgit = Versioningit.from_project_dir(PROJECT_ROOT)
        report = vgit.run(write=True, fallback=True)
    except NotVersioningitError:
        log.info("versioningit not enabled in pyproject.toml; doing nothing")
        return
    except (NotSdistError, NoTagError):
        raise RuntimeError(
            "\nversioningit could not find a version for the project in"
            f" {PROJECT_ROOT}!\n\n"
            "You may be installing from a shallow clone, in which case you"
            " need to unshallow it first.\n\n"
            "Alternatively, you may be installing from a Git archive, which is"
            " not supported by default.  Install from a git+https://... URL"
            " instead.\n\n"
        )
    dist.metadata.version = report.version
    if isinstance(report, Report):
        dist._versioningit_template_fields = (  # type: ignore[attr-defined]
            report.template_fields
        )



from __future__ import annotations
from dataclasses import Field, dataclass, field, fields
from pathlib import Path
import sys
from typing import Any, Optional
from .errors import ConfigError, NotVersioningitError
from .logging import warn_extra_fields
from .methods import (
    CallableSpec,
    CustomMethodSpec,
    EntryPointSpec,
    MethodSpec,
    VersioningitMethod,
)
from .util import optional_str_guard

if sys.version_info[:2] >= (3, 11):
    from tomllib import load as toml_load
else:
    from tomli import load as toml_load


@dataclass
class ConfigSection:
    """A parsed method subtable of the `versioningit` configuration"""

    #: Specification of the method for this subtable
    method_spec: MethodSpec

    #: Additional parameters to pass to the method
    params: dict[str, Any]

    def load(self, project_dir: str | Path) -> VersioningitMethod:
        """Loads the method and returns a `VersioningitMethod`"""
        return VersioningitMethod(self.method_spec.load(project_dir), self.params)


@dataclass
class Config:
    """Parsed `versioningit` configuration"""

    #: Parsed ``vcs`` subtable
    vcs: ConfigSection = field(metadata={"default_entry_point": "git"})

    #: Parsed ``tag2version`` subtable
    tag2version: ConfigSection = field(metadata={"default_entry_point": "basic"})

    #: Parsed ``next-version`` subtable
    next_version: ConfigSection = field(metadata={"default_entry_point": "minor"})

    #: Parsed ``format`` subtable
    format: ConfigSection = field(metadata={"default_entry_point": "basic"})

    #: Parsed ``template-fields`` subtable
    template_fields: ConfigSection = field(metadata={"default_entry_point": "basic"})

    #: Parsed ``write`` subtable
    write: Optional[ConfigSection] = field(
        default=None, metadata={"default_entry_point": "basic", "optional": True}
    )

    #: Parsed ``onbuild`` subtable
    onbuild: Optional[ConfigSection] = field(
        default=None,
        metadata={"default_entry_point": "replace-version", "optional": True},
    )

    #: The ``default-version`` setting
    default_version: Optional[str] = None

    @classmethod
    def parse_toml_file(cls, filepath: str | Path) -> Config:
        """
        Parse the ``[tool.versioningit]`` table in the given TOML file

        :raises NotVersioningitError:
            if the file does not contain a ``[tool.versioningit]`` table
        :raises ConfigError:
            if the ``tool.versioningit`` key or any of its subfields are not of
            the correct type
        """
        with open(filepath, "rb") as fp:
            data = toml_load(fp).get("tool", {}).get("versioningit")
        if data is None:
            raise NotVersioningitError("versioningit not enabled in pyproject.toml")
        return cls.parse_obj(data)

    @classmethod
    def parse_obj(cls, obj: Any) -> Config:
        """
        Parse a raw Python configuration structure

        :raises ConfigError:
            - if ``obj`` is not a `dict`
            - if ``default-version`` or any of the subtable or ``method``
              fields are not of the correct type
        """
        if not isinstance(obj, dict):
            raise ConfigError("tool.versioningit must be a table")
        default_version = optional_str_guard(
            obj.pop("default-version", None), "tool.versioningit.default-version"
        )
        sections: dict[str, Optional[ConfigSection]] = {}
        for f in fields(cls):
            if not f.metadata:
                continue
            sections[f.name] = cls.parse_section(f, obj.pop(attr2key(f.name), None))
        warn_extra_fields(
            obj,
            "tool.versioningit",
            [attr2key(f.name) for f in fields(cls)],
        )
        return cls(
            default_version=default_version, **sections  # type: ignore[arg-type]
        )

    @staticmethod
    def parse_section(f: Field, obj: Any) -> Optional["ConfigSection"]:
        """
        Parse a ``tool.versioniningit.STEP`` field according to the metadata in
        the given `dataclasses.Field`, which must consist of the following
        items:

        ``default_entry_point`` : string
            The name of the default method to use for the step if one is not
            specified in the configuration

        ``optional`` : bool
            If true, an absent/`None` step field will be converted to `None`
            instead of to a section with the default method and no user
            parameters.  If not specified, defaults to false.

        :raises ConfigError:
            - if ``obj`` is not `None`, a callable, a string, or a `dict`
            - if any of the ``method`` fields are not of the correct type
        """
        if obj is None:
            if f.metadata.get("optional"):
                return None
            else:
                obj = f.metadata["default_entry_point"]
        if isinstance(obj, str):
            method_spec = Config.parse_method_spec(f, obj)
            return ConfigSection(method_spec, {})
        elif callable(obj):
            method_spec = CallableSpec(obj)
            return ConfigSection(method_spec, {})
        elif isinstance(obj, dict):
            if "method" not in obj and "module" in obj and "value" in obj:
                method_spec = Config.parse_method_spec(f, obj)
                return ConfigSection(method_spec, {})
            method_spec = Config.parse_method_spec(f, obj.pop("method", None))
            return ConfigSection(method_spec, obj)
        else:
            raise ConfigError(
                f"tool.versioningit.{attr2key(f.name)} must be a string or table"
            )

    @staticmethod
    def parse_method_spec(f: Field, method: Any) -> MethodSpec:
        """
        Parse a ``method`` field according to the metadata in the given
        `dataclasses.Field` (see `parse_section()`)

        :raises ConfigError:
            - if ``method`` is not `None`, a string, a callable, or a `dict`
            - if ``method`` is a `dict` without a ``module`` or ``value`` key
            - if any of the fields in ``method`` are not of the correct type
        """
        key = attr2key(f.name)
        if method is None:
            method = f.metadata["default_entry_point"]
        if isinstance(method, str):
            return EntryPointSpec(group=f"versioningit.{f.name}", name=method)
        elif callable(method):
            return CallableSpec(method)
        elif isinstance(method, dict):
            module = method.pop("module", None)
            if not isinstance(module, str):
                raise ConfigError(
                    f"tool.versioningit.{key}.method.module is required and"
                    " must be a string"
                )
            value = method.pop("value", None)
            if not isinstance(value, str):
                raise ConfigError(
                    f"tool.versioningit.{key}.method.value is required and"
                    " must be a string"
                )
            module_dir = method.pop("module-dir", None)
            if module_dir is not None and not isinstance(module_dir, str):
                raise ConfigError(
                    f"tool.versioningit.{key}.method.module-dir must be a string"
                )
            warn_extra_fields(
                method,
                f"tool.versioningit.{key}.method",
                ["module", "value", "module-dir"],
            )
            return CustomMethodSpec(module, value, module_dir)
        else:
            raise ConfigError(
                f"tool.versioningit.{key}.method must be a string or table"
            )


def attr2key(name: str) -> str:
    return name.replace("_", "-")



from __future__ import annotations
from dataclasses import dataclass
import re
from typing import Any, Optional
from packaging.version import Version
from .errors import InvalidVersionError
from .logging import warn_extra_fields


@dataclass
class BasicVersion:
    """A version consisting of just an optional epoch and a release segment"""

    #: The epoch (Zero equals a lack of an explicit epoch)
    epoch: int

    #: The integer values of the components of the release segment
    release: list[int]

    @classmethod
    def parse(cls, version: str) -> BasicVersion:
        """
        Parse the initial epoch and release segment from a version string and
        discard any other trailing characters

        :raises InvalidVersionError: if ``version`` cannot be parsed
        """
        m = re.match(
            r"v?(?:(?P<epoch>[0-9]+)!)?(?P<release>[0-9]+(?:\.[0-9]+)*)(?!!)", version
        )
        if not m:
            raise InvalidVersionError(f"Cannot parse version {version!r}")
        sepoch = m["epoch"]
        if sepoch is None:
            epoch = 0  # type: ignore[unreachable]
        else:
            assert isinstance(sepoch, str)
            epoch = int(sepoch)
        release = m["release"]
        assert isinstance(release, str)
        return cls(epoch, list(map(int, release.split("."))))

    def __str__(self) -> str:
        """Convert the `BasicVersion` to a string"""
        s = ""
        if self.epoch > 0:
            s += f"{self.epoch}!"
        s += ".".join(map(str, self.release))
        return s


def next_minor_version(
    *,
    version: str,
    branch: Optional[str],  # noqa: U100
    params: dict[str, Any],
) -> str:
    """Implements the ``"minor"`` ``next-version`` method"""
    warn_extra_fields(params, "tool.versioningit.next-version")
    bv = BasicVersion.parse(version)
    bv.release = (bv.release + [0, 0])[:2]
    bv.release[1] += 1
    bv.release.append(0)
    return str(bv)


def next_smallest_version(
    *,
    version: str,
    branch: Optional[str],  # noqa: U100
    params: dict[str, Any],
) -> str:
    """Implements the ``"smallest"`` ``next-version`` method"""
    warn_extra_fields(params, "tool.versioningit.next-version")
    bv = BasicVersion.parse(version)
    bv.release[-1] += 1
    return str(bv)


def null_next_version(
    *,
    version: str,
    branch: Optional[str],  # noqa: U100
    params: dict[str, Any],
) -> str:
    """Implements the ``"null"`` ``next-version`` method"""
    warn_extra_fields(params, "tool.versioningit.next-version")
    return version


def next_minor_release_version(
    *,
    version: str,
    branch: Optional[str],  # noqa: U100
    params: dict[str, Any],
) -> str:
    """
    Implements the ``"minor-release"`` ``next-version`` method.

    If ``version`` is a prerelease version, returns the base version.
    Otherwise, returns the next minor version after the base version.
    """
    warn_extra_fields(params, "tool.versioningit.next-version")
    try:
        v = Version(version)
    except ValueError:
        raise InvalidVersionError(f"Cannot parse version {version!r}")
    if v.is_prerelease:
        return str(v.base_version)
    vs = list(v.release) + [0, 0]
    vs[1] += 1
    vs[2:] = [0]
    s = ".".join(map(str, vs))
    if v.epoch:
        s = f"{v.epoch}!{s}"
    return s


def next_smallest_release_version(
    *,
    version: str,
    branch: Optional[str],  # noqa: U100
    params: dict[str, Any],
) -> str:
    """
    Implements the ``"smallest-release"`` ``next-version`` method.

    If ``version`` is a prerelease version, returns the base version.
    Otherwise, returns the next smallest version after the base version.
    """
    warn_extra_fields(params, "tool.versioningit.next-version")
    try:
        v = Version(version)
    except ValueError:
        raise InvalidVersionError(f"Cannot parse version {version!r}")
    if v.is_prerelease:
        return str(v.base_version)
    vs = list(v.release)
    vs[-1] += 1
    s = ".".join(map(str, vs))
    if v.epoch:
        s = f"{v.epoch}!{s}"
    return s



"""
Versioning It with your Version In Git

``versioningit`` is yet another setuptools plugin for automatically determining
your package's version based on your Git repository's tags.  Unlike others, it
allows easy customization of the version format and even lets you easily
override the separate functions used for version extraction & calculation.

**Features:**

- Installed & configured through :pep:`518`'s ``pyproject.toml``

- Supports Git, modern Git archives, and Mercurial

- Formatting of the final version uses format template strings, with fields for
  basic VCS information and separate template strings for distanced vs. dirty
  vs. distanced-and-dirty repository states

- Can optionally write the final version to a file for loading at runtime

- Provides custom setuptools commands for inserting the final version into a
  source file at build time

- The individual methods for VCS querying, tag-to-version calculation, version
  bumping, version formatting, and writing the version to a file can all be
  customized using either functions defined alongside one's project code or via
  publicly-distributed entry points

- Can alternatively be used as a library for use in ``setup.py`` or the like,
  in case you don't want to or can't configure it via ``pyproject.toml``

- The only thing it does is calculate your version and optionally write it to a
  file; there's no overriding of your sdist contents based on what's in your
  Git repository, especially not without a way to turn it off, because that
  would just be rude.

Visit <https://github.com/jwodder/versioningit> or
<https://versioningit.rtfd.io> for more information.
"""

__version__ = "2.1.0"
__author__ = "John Thorvald Wodder II"
__author_email__ = "versioningit@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/versioningit"

from .cmdclasses import get_cmdclasses
from .core import (
    FallbackReport,
    Report,
    VCSDescription,
    Versioningit,
    get_next_version,
    get_template_fields_from_distribution,
    get_version,
    get_version_from_pkg_info,
    run_onbuild,
)
from .errors import (
    ConfigError,
    Error,
    InvalidTagError,
    InvalidVersionError,
    MethodError,
    NoTagError,
    NotSdistError,
    NotVCSError,
    NotVersioningitError,
)

__all__ = [
    "ConfigError",
    "Error",
    "FallbackReport",
    "InvalidTagError",
    "InvalidVersionError",
    "MethodError",
    "NoTagError",
    "NotSdistError",
    "NotVCSError",
    "NotVersioningitError",
    "Report",
    "VCSDescription",
    "Versioningit",
    "get_cmdclasses",
    "get_next_version",
    "get_template_fields_from_distribution",
    "get_version",
    "get_version_from_pkg_info",
    "run_onbuild",
]



from __future__ import annotations
from difflib import get_close_matches
import logging
import os
from typing import Iterable, Optional
from packaging.version import Version

log = logging.getLogger(__package__)


def get_env_loglevel() -> Optional[int]:
    """
    Return the logging level specified in the :envvar:`VERSIONINGIT_LOG_LEVEL`
    environment variable, if any
    """
    try:
        return parse_log_level(os.environ["VERSIONINGIT_LOG_LEVEL"])
    except (KeyError, ValueError):
        return None


def init_logging(level: Optional[int] = None) -> None:
    """
    Configure the `versioningit` logger and set its level to ``level``.  If
    ``level`` is not specified, the value is taken from the
    :envvar:`VERSIONINGIT_LOG_LEVEL` environment variable, with a default of
    ``WARNING``.
    """
    if log.handlers:
        return
    if level is None:
        level = get_env_loglevel()
    if level is None:
        level = logging.WARNING
    log.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("[%(levelname)-8s] %(name)s: %(message)s"))
    log.addHandler(handler)
    log.propagate = False


def parse_log_level(level: str) -> int:
    """
    Convert a log level name (case-insensitive) or number to its numeric value
    """
    try:
        return int(level)
    except ValueError:
        levelup = level.upper()
        if levelup in {"CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"}:
            ll = getattr(logging, levelup)
            assert isinstance(ll, int)
            return ll
        else:
            raise ValueError(f"Invalid log level: {level!r}")


def warn_extra_fields(
    params: dict, fieldname: str, valid: Optional[list[str]] = None
) -> None:
    """
    For each key in ``params``, emit a log message indicating that the given
    parameter is ignored, along with a "Did you mean?" message if ``valid`` is
    set and the key resembles any elements of ``valid``.  ``fieldname`` is the
    name of the table in which ``params`` were found.
    """
    for p in params.keys():
        suggestions = didyoumean(p, valid) if valid is not None else ""
        log.warning("Ignoring unknown parameter %r in %s%s", p, fieldname, suggestions)


def warn_bad_version(version: str, desc: str) -> None:
    """
    If ``version`` is not :pep:`440`-compliant, log a warning.  ``desc`` is a
    description of the version's provenance.
    """
    try:
        Version(version)
    except ValueError:
        log.warning("%s %r is not PEP 440-compliant", desc, version)


def didyoumean(mistake: str, valid: Iterable[str]) -> str:
    """
    If ``mistake`` resembles any elements of ``valid``, return a "Did you
    mean?" message enclosed in parentheses and starting with a space.  If
    ``mistake`` is not similar to any elements of ``valid``, return an empty
    string.
    """
    candidates = get_close_matches(mistake, valid)
    if candidates:
        return " (Did you mean:" + "".join(f" {c}?" for c in candidates) + ")"
    else:
        return ""



class Error(Exception):
    """Base class of all ``versioningit``-specific errors"""

    pass


class ConfigError(Error, ValueError):
    """
    Raised when the ``versioningit`` configuration contain invalid settings
    """

    pass


class MethodError(Error):
    """Raised when a method is invalid or returns an invalid value"""

    pass


class NotVersioningitError(Error):
    """
    Raised when ``versioningit`` is used on a project that does not have
    ``versioningit`` enabled
    """

    pass


class NotSdistError(Error):
    """
    Raised when attempting to read a :file:`PKG-INFO` file from a directory
    that doesn't have one
    """

    pass


class NotVCSError(Error):
    """
    Raised when ``versioningit`` is run in a directory that is not under
    version control or when the relevant VCS program is not installed
    """

    pass


class NoTagError(Error):
    """Raised when a tag cannot be found in version control"""

    pass


class InvalidTagError(Error, ValueError):
    """
    Raised by ``tag2version`` methods when passed a tag that they cannot work
    with
    """

    pass


class InvalidVersionError(Error, ValueError):
    """
    Raised by ``next-version`` and ``template-fields`` methods when passed a
    version that they cannot work with
    """

    pass



from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from importlib import import_module
import os.path
from pathlib import Path
import sys
from typing import Any, Optional, cast
from .errors import ConfigError, MethodError
from .logging import didyoumean, log

if sys.version_info[:2] >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points


# Call `entry_points()` only once and save the results for a speedup.  See
# <https://github.com/jwodder/versioningit/pull/7>.
ENTRY_POINTS = entry_points()


class MethodSpec(ABC):
    """
    An abstract base class for method specifications parsed from `versioningit`
    configurations
    """

    @abstractmethod
    def load(self, project_dir: str | Path) -> Callable:
        """
        Load & return the callable specified by the `MethodSpec`.
        ``project_dir`` is provided in case the method needs to load anything
        from the project directory itself.
        """
        ...


@dataclass
class EntryPointSpec(MethodSpec):
    """
    A parsed method specification identifying a Python packaging entry point
    """

    #: The name of the group in which to look up the entry point
    group: str

    #: The name of the entry point
    name: str

    def load(self, _project_dir: str | Path) -> Callable:
        """
        Loads & returns the entry point

        :raises ConfigError: if no such entry point exists
        :raises MethodError: if the loaded entry point is not a callable
        """
        log.debug("Loading entry point %r in group %s", self.name, self.group)
        eps = list(ENTRY_POINTS.select(group=self.group, name=self.name))
        if len(eps) == 0:
            valid = [ep.name for ep in ENTRY_POINTS.select(group=self.group)]
            raise ConfigError(
                f"{self.group} entry point {self.name!r} not"
                f" found{didyoumean(self.name, valid)}"
            )
        elif len(eps) > 1:
            raise ConfigError(
                "Packaging conflict!  Multiple entry points named"
                f" {self.name!r} registered in group {self.group}"
            )
        else:
            ep = eps[0]
        c = ep.load()
        if not callable(c):
            raise MethodError(
                f"{self.group} entry point {self.name!r} did not resolve to a"
                " callable object"
            )
        return cast(Callable, c)


@dataclass
class CustomMethodSpec(MethodSpec):
    """
    A parsed method specification identifying a callable in a local Python
    module
    """

    #: The dotted name of the module containing the callable
    module: str

    #: The name of the callable object within the module
    value: str

    #: The directory in which the module is located; defaults to
    #: ``project_dir``
    module_dir: Optional[str]

    def load(self, project_dir: str | Path) -> Callable:
        """
        Loads the module and returns the callable

        :raises MethodError: if the object is not actually a callable
        """
        if self.module_dir is not None:
            modpath = os.path.join(project_dir, self.module_dir)
        else:
            modpath = str(project_dir)
        log.debug("Prepending %s to sys.path", modpath)
        sys.path.insert(0, modpath)
        try:
            log.debug("Importing %s from %s", self.value, self.module)
            obj = import_module(self.module)
        finally:
            log.debug("Removing %s from sys.path", modpath)
            with suppress(ValueError):
                sys.path.remove(modpath)
        for attr in self.value.split("."):
            obj = getattr(obj, attr)
        if not callable(obj):
            raise MethodError(
                f"Custom method '{self.module}:{self.value}' did not resolve"
                " to a callable object"
            )
        return cast(Callable, obj)


@dataclass
class CallableSpec(MethodSpec):
    """
    A parsed method specification identifying a callable by the callable itself
    """

    #: The callable
    func: Callable

    def load(self, _project_dir: str | Path) -> Callable:
        """Return the callable"""
        return self.func


@dataclass
class VersioningitMethod:
    """
    A loaded `versioningit` method and the user-supplied parameters to pass to
    it
    """

    #: The loaded method
    method: Callable

    #: User-supplied parameters obtained from the original configuration
    params: dict[str, Any]

    def __call__(self, **kwargs: Any) -> Any:
        """
        Invokes the method with the given keyword arguments and the
        user-supplied parameters
        """
        return self.method(params=self.params, **kwargs)



from __future__ import annotations
from dataclasses import dataclass
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional
from .config import Config
from .errors import Error, MethodError, NotSdistError, NotVCSError, NotVersioningitError
from .logging import log, warn_bad_version
from .methods import VersioningitMethod
from .util import is_sdist, parse_version_from_metadata

if TYPE_CHECKING:
    from setuptools import Distribution


@dataclass
class VCSDescription:
    """A description of the state of a version control repository"""

    #: The name of the most recent tag in the repository (possibly after
    #: applying any match or exclusion rules based on user parameters) from
    #: which the current repository state is descended
    tag: str

    #: The relationship of the repository's current state to the tag.  If the
    #: repository state is exactly the tagged state, this field should equal
    #: ``"exact"``; otherwise, it will be a string that will be used as a key
    #: in the ``[tool.versioningit.format]`` subtable.  Recommended values are
    #: ``"distance"``, ``"dirty"``, and ``"distance-dirty"``.
    state: str

    #: The name of the repository's current branch, or `None` if it cannot be
    #: determined or does not apply
    branch: Optional[str]

    #: A `dict` of additional information about the repository state to make
    #: available to the ``format`` method.  Custom ``vcs`` methods are advised
    #: to adhere closely to the set of fields used by the built-in methods.
    fields: dict[str, Any]


@dataclass
class Report:
    """
    .. versionadded:: 2.0.0

    A report of the intermediate & final values calculated during a
    ``versioningit`` run
    """

    #: The final version
    version: str

    #: A description of the state of the version control repository; `None` if
    #: the "vcs" step failed
    description: Optional[VCSDescription]

    #: A version string extracted from the VCS tag; `None` if the "tag2version"
    #: step or a previous step failed
    base_version: Optional[str]

    #: A "next version" calculated by the "next-version" step; `None` if the
    #: step or a previous one failed
    next_version: Optional[str]

    #: A `dict` of fields for use in templating by the "write" and "onbuild"
    #: steps
    template_fields: dict[str, Any]

    #: `True` iff an error occurred during version calculation, causing a
    #: ``default-version`` setting to be used
    using_default_version: bool


@dataclass
class FallbackReport:
    """
    .. versionadded:: 2.0.0

    A report of the version extracted from a :file:`PKG-INFO` file in an sdist
    """

    #: The version
    version: str


@dataclass
class Versioningit:
    """
    A class for getting a version-controlled project's current version based on
    its most recent tag and the difference therefrom
    """

    #: The path to the root of the project directory (usually the location of a
    #: :file:`pyproject.toml` file)
    #:
    #: :meta private:
    project_dir: Path

    #: The default version, if any, to use if an error occurs
    #:
    #: :meta private:
    default_version: Optional[str]

    #: The method to call for the ``vcs`` step
    #:
    #: :meta private:
    vcs: VersioningitMethod

    #: The method to call for the ``tag2version`` step
    #:
    #: :meta private:
    tag2version: VersioningitMethod

    #: The method to call for the ``next-version`` step
    #:
    #: :meta private:
    next_version: VersioningitMethod

    #: The method to call for the ``format`` step
    #:
    #: :meta private:
    format: VersioningitMethod

    #: The method to call for the ``template-fields`` step
    #:
    #: :meta private:
    template_fields: VersioningitMethod

    #: The method to call for the ``write`` step
    #:
    #: :meta private:
    write: Optional[VersioningitMethod]

    #: The method to call for the ``onbuild`` step
    #:
    #: :meta private:
    onbuild: Optional[VersioningitMethod]

    @classmethod
    def from_project_dir(
        cls, project_dir: str | Path = os.curdir, config: Optional[dict] = None
    ) -> Versioningit:
        """
        Construct a `Versioningit` object for the project rooted at
        ``project_dir`` (default: the current directory).

        If ``config`` is `None`, then ``project_dir`` must contain a
        :file:`pyproject.toml` file containing a ``[tool.versioningit]`` table;
        if it does not, a `NotVersioningitError` is raised.  If ``config`` is
        not `None`, then any :file:`pyproject.toml` file in ``project_dir``
        will be ignored, and the configuration will be taken from ``config``
        instead.  See ":ref:`config_dict`".

        :raises NotVersioningitError:
            - if ``config`` is `None` and ``project_dir`` does not contain a
              :file:`pyproject.toml` file
            - if ``config`` is `None` and the :file:`pyproject.toml` file does
              not contain a ``[tool.versioningit]`` table
        :raises ConfigError:
            if the ``tool.versioningit`` key, ``config``, or any subfields are
            not of the correct type
        """
        if config is None:
            try:
                cfg = Config.parse_toml_file(Path(project_dir, "pyproject.toml"))
            except FileNotFoundError:
                raise NotVersioningitError(f"No pyproject.toml file in {project_dir}")
        else:
            cfg = Config.parse_obj(config)
        return cls.from_config(project_dir, cfg)

    @classmethod
    def from_config(cls, project_dir: str | Path, config: Config) -> Versioningit:
        """
        Construct a `Versioningit` object from a parsed configuration object

        :meta private:
        """
        project_dir = Path(project_dir)
        return cls(
            project_dir=project_dir,
            default_version=config.default_version,
            vcs=config.vcs.load(project_dir),
            tag2version=config.tag2version.load(project_dir),
            next_version=config.next_version.load(project_dir),
            format=config.format.load(project_dir),
            template_fields=config.template_fields.load(project_dir),
            write=config.write.load(project_dir) if config.write is not None else None,
            onbuild=config.onbuild.load(project_dir)
            if config.onbuild is not None
            else None,
        )

    def get_version(self, write: bool = False, fallback: bool = True) -> str:
        """
        Determine the version for the project.

        If ``write`` is true, then the file specified in the
        ``[tool.versioningit.write]`` subtable, if any, will be updated.

        If ``fallback`` is true, then if ``project_dir`` is not under version
        control (or if the VCS executable is not installed), ``versioningit``
        will assume that the directory is an unpacked sdist and will read the
        version from the :file:`PKG-INFO` file.

        .. versionchanged:: 2.0.0

            ``write`` and ``fallback`` arguments added

        :raises NotVCSError:
            if ``fallback`` is false and ``project_dir`` is not under version
            control
        :raises NotSdistError:
            if ``fallback`` is true, ``project_dir`` is not under version
            control, and there is no :file:`PKG-INFO` file in ``project_dir``
        :raises ConfigError:
            if any of the values in ``config`` are not of the correct type
        :raises MethodError: if a method returns a value of the wrong type
        """
        return self.run(write=write, fallback=fallback).version

    def run(
        self, write: bool = False, fallback: bool = True
    ) -> Report | FallbackReport:
        """
        .. versionadded:: 2.0.0

        Run all of the steps for the project — aside from "onbuild" and,
        optionally, "write" — and return an object containing the final version
        and intermediate values.

        If ``write`` is true, then the file specified in the
        ``[tool.versioningit.write]`` subtable, if any, will be updated.

        If ``fallback`` is true, then if ``project_dir`` is not under version
        control (or if the VCS executable is not installed), ``versioningit``
        will assume that the directory is an unpacked sdist and will read the
        version from the :file:`PKG-INFO` file, returning a `FallbackReport`
        instance instead of a `Report`.

        :raises NotVCSError:
            if ``fallback`` is false and ``project_dir`` is not under version
            control
        :raises NotSdistError:
            if ``fallback`` is true, ``project_dir`` is not under version
            control, and there is no :file:`PKG-INFO` file in ``project_dir``
        :raises ConfigError:
            if any of the values in ``config`` are not of the correct type
        :raises MethodError: if a method returns a value of the wrong type
        """
        description: Optional[VCSDescription] = None
        base_version: Optional[str] = None
        next_version: Optional[str] = None
        using_default_version: bool = False
        try:
            description = self.do_vcs()
            base_version = self.do_tag2version(description.tag)
            next_version = self.do_next_version(base_version, description.branch)
            if description.state == "exact":
                log.info("Tag is exact match; returning extracted version")
                version = base_version
            else:
                log.info("VCS state is %r; formatting version", description.state)
                version = self.do_format(
                    description=description,
                    base_version=base_version,
                    next_version=next_version,
                )
            log.info("Final version: %s", version)
        except Error as e:
            if (
                isinstance(e, NotVCSError)
                and fallback
                and (is_sdist(self.project_dir) or self.default_version is None)
            ):
                log.info("Could not get VCS data from %s: %s", self.project_dir, str(e))
                log.info("Falling back to reading from PKG-INFO")
                return FallbackReport(
                    version=get_version_from_pkg_info(self.project_dir)
                )
            if self.default_version is not None:
                log.error("%s: %s", type(e).__name__, str(e))
                log.info("Falling back to tool.versioningit.default-version")
                version = self.default_version
                using_default_version = True
            else:
                raise
        except Exception:  # pragma: no cover
            if self.default_version is not None:
                log.exception("An unexpected error occurred:")
                log.info("Falling back to tool.versioningit.default-version")
                version = self.default_version
                using_default_version = True
            else:
                raise
        warn_bad_version(version, "Final version")
        template_fields = self.do_template_fields(
            version=version,
            description=description,
            base_version=base_version,
            next_version=next_version,
        )
        if write:
            self.do_write(template_fields)
        return Report(
            version=version,
            description=description,
            base_version=base_version,
            next_version=next_version,
            template_fields=template_fields,
            using_default_version=using_default_version,
        )

    def do_vcs(self) -> VCSDescription:
        """
        Run the ``vcs`` step

        :raises MethodError: if the method does not return a `VCSDescription`
        """
        description = self.vcs(project_dir=self.project_dir)
        if not isinstance(description, VCSDescription):
            raise MethodError(
                f"vcs method returned {description!r} instead of a VCSDescription"
            )
        log.info("vcs returned tag %s", description.tag)
        log.debug("vcs state: %s", description.state)
        log.debug("vcs branch: %s", description.branch)
        log.debug("vcs fields: %r", description.fields)
        return description

    def do_tag2version(self, tag: str) -> str:
        """
        Run the ``tag2version`` step

        :raises MethodError: if the method does not return a `str`
        """
        version = self.tag2version(tag=tag)
        if not isinstance(version, str):
            raise MethodError(
                f"tag2version method returned {version!r} instead of a string"
            )
        log.info("tag2version returned version %s", version)
        warn_bad_version(version, "Version extracted from tag")
        return version

    def do_next_version(self, version: str, branch: Optional[str]) -> str:
        """
        Run the ``next-version`` step

        :raises MethodError: if the method does not return a `str`
        """
        next_version = self.next_version(version=version, branch=branch)
        if not isinstance(next_version, str):
            raise MethodError(
                f"next-version method returned {next_version!r} instead of a string"
            )
        log.info("next-version returned version %s", next_version)
        warn_bad_version(next_version, "Calculated next version")
        return next_version

    def do_format(
        self, description: VCSDescription, base_version: str, next_version: str
    ) -> str:
        """
        Run the ``format`` step

        .. versionchanged:: 2.0.0

            The ``version`` argument was renamed to ``base_version``.

        :raises MethodError: if the method does not return a `str`
        """
        new_version = self.format(
            description=description,
            base_version=base_version,
            next_version=next_version,
        )
        if not isinstance(new_version, str):
            raise MethodError(
                f"format method returned {new_version!r} instead of a string"
            )
        return new_version

    def do_template_fields(
        self,
        version: str,
        description: Optional[VCSDescription],
        base_version: Optional[str],
        next_version: Optional[str],
    ) -> dict:
        """
        .. versionadded:: 2.0.0

        Run the ``template_fields`` step

        :raises MethodError: if the method does not return a `dict`
        """
        fields = self.template_fields(
            version=version,
            description=description,
            base_version=base_version,
            next_version=next_version,
        )
        if not isinstance(fields, dict):
            raise MethodError(
                f"template-fields method returned {fields!r} instead of a dict"
            )
        log.debug("Template fields available to `write` and `onbuild`: %r", fields)
        return fields

    def do_write(self, template_fields: dict[str, Any]) -> None:
        """
        Run the ``write`` step

        .. versionchanged:: 2.0.0

            ``version`` argument replaced with ``template_fields``
        """
        if self.write is not None:
            self.write(project_dir=self.project_dir, template_fields=template_fields)
        else:
            log.info("'write' step not configured; not writing anything")

    def do_onbuild(
        self,
        build_dir: str | Path,
        is_source: bool,
        template_fields: dict[str, Any],
    ) -> None:
        """
        .. versionadded:: 1.1.0

        Run the ``onbuild`` step

        .. versionchanged:: 2.0.0

            ``version`` argument replaced with ``template_fields``
        """
        if self.onbuild is not None:
            self.onbuild(
                build_dir=build_dir,
                is_source=is_source,
                template_fields=template_fields,
            )
        else:
            log.info("'onbuild' step not configured; not doing anything")


def get_version(
    project_dir: str | Path = os.curdir,
    config: Optional[dict] = None,
    write: bool = False,
    fallback: bool = True,
) -> str:
    """
    Determine the version for the project at ``project_dir``.

    If ``config`` is `None`, then ``project_dir`` must contain a
    :file:`pyproject.toml` file containing a ``[tool.versioningit]`` table; if
    it does not, a `NotVersioningitError` is raised.  If ``config`` is not
    `None`, then any :file:`pyproject.toml` file in ``project_dir`` will be
    ignored, and the configuration will be taken from ``config`` instead; see
    ":ref:`config_dict`".

    If ``write`` is true, then the file specified in the
    ``[tool.versioningit.write]`` subtable, if any, will be updated.

    If ``fallback`` is true, then if ``project_dir`` is not under version
    control (or if the VCS executable is not installed), ``versioningit`` will
    assume that the directory is an unpacked sdist and will read the version
    from the :file:`PKG-INFO` file.

    :raises NotVCSError:
        if ``fallback`` is false and ``project_dir`` is not under version
        control
    :raises NotSdistError:
        if ``fallback`` is true, ``project_dir`` is not under version control,
        and there is no :file:`PKG-INFO` file in ``project_dir``
    :raises NotVersioningitError:
        - if ``config`` is `None` and ``project_dir`` does not contain a
          :file:`pyproject.toml` file
        - if the :file:`pyproject.toml` file does not contain a
          ``[tool.versioningit]`` table
    :raises ConfigError:
        if any of the values in ``config`` are not of the correct type
    :raises MethodError: if a method returns a value of the wrong type
    """
    vgit = Versioningit.from_project_dir(project_dir, config)
    return vgit.get_version(write=write, fallback=fallback)


def get_next_version(
    project_dir: str | Path = os.curdir, config: Optional[dict] = None
) -> str:
    """
    .. versionadded:: 0.3.0

    Determine the next version after the current VCS-tagged version for
    ``project_dir``.

    If ``config`` is `None`, then ``project_dir`` must contain a
    :file:`pyproject.toml` file containing a ``[tool.versioningit]`` table; if
    it does not, a `NotVersioningitError` is raised.  If ``config`` is not
    `None`, then any :file:`pyproject.toml` file in ``project_dir`` will be
    ignored, and the configuration will be taken from ``config`` instead; see
    ":ref:`config_dict`".

    :raises NotVCSError:
        if ``project_dir`` is not under version control
    :raises NotVersioningitError:
        - if ``config`` is `None` and ``project_dir`` does not contain a
          :file:`pyproject.toml` file
        - if the :file:`pyproject.toml` file does not contain a
          ``[tool.versioningit]`` table
    :raises ConfigError:
        if any of the values in ``config`` are not of the correct type
    :raises MethodError: if a method returns a value of the wrong type
    """
    vgit = Versioningit.from_project_dir(project_dir, config)
    description = vgit.do_vcs()
    tag_version = vgit.do_tag2version(description.tag)
    next_version = vgit.do_next_version(tag_version, description.branch)
    return next_version


def get_version_from_pkg_info(project_dir: str | Path) -> str:
    """
    Return the :mailheader:`Version` field from the :file:`PKG-INFO` file in
    ``project_dir``

    :raises NotSdistError: if there is no :file:`PKG-INFO` file
    :raises ValueError:
        if the :file:`PKG-INFO` file does not contain a :mailheader:`Version`
        field
    """
    try:
        return parse_version_from_metadata(
            Path(project_dir, "PKG-INFO").read_text(encoding="utf-8")
        )
    except FileNotFoundError:
        raise NotSdistError(f"{project_dir} does not contain a PKG-INFO file")


def run_onbuild(
    *,
    build_dir: str | Path,
    is_source: bool,
    template_fields: dict[str, Any],
    project_dir: str | Path = os.curdir,
    config: Optional[dict] = None,
) -> None:
    """
    .. versionadded:: 1.1.0

    Run the ``onbuild`` step for the given project.

    This function is intended to be used by custom setuptools commands that are
    used in place of ``versioningit``'s custom commands but still need to be
    able to run the ``onbuild`` step.  The ``template_fields`` value can be
    obtained by passing the command's ``distribution`` attribute to
    `get_template_fields_from_distribution()`; if this returns `None`, then we
    are building from an sdist, and `run_onbuild()` should not be called.

    If ``config`` is `None`, then ``project_dir`` must contain a
    :file:`pyproject.toml` file containing a ``[tool.versioningit]`` table; if
    it does not, a `NotVersioningitError` is raised.  If ``config`` is not
    `None`, then any :file:`pyproject.toml` file in ``project_dir`` will be
    ignored, and the configuration will be taken from ``config`` instead; see
    ":ref:`config_dict`".

    .. versionchanged:: 2.0.0

        ``version`` argument replaced with ``template_fields``

    :param build_dir: The directory containing the in-progress build
    :param is_source:
        Set to `True` if building an sdist or other artifact that preserves
        source paths, `False` if building a wheel or other artifact that uses
        installation paths
    :param template_fields: A `dict` of fields to be used when templating
    :raises NotVersioningitError:
        - if ``config`` is `None` and ``project_dir`` does not contain a
          :file:`pyproject.toml` file
        - if the :file:`pyproject.toml` file does not contain a
          ``[tool.versioningit]`` table
    :raises ConfigError:
        if any of the values in ``config`` are not of the correct type
    :raises MethodError: if a method returns a value of the wrong type
    """
    vgit = Versioningit.from_project_dir(project_dir, config)
    vgit.do_onbuild(
        build_dir=build_dir, is_source=is_source, template_fields=template_fields
    )


def get_template_fields_from_distribution(
    dist: Distribution,
) -> Optional[dict[str, Any]]:
    """
    Extract the template fields (calculated by the "template-fields" step) that
    were stashed on the `setuptools.Distribution` by ``versioningit``'s
    setuptools hook, for passing to the "onbuild" step.  If setuptools is
    building from an sdist instead of a repository, no template fields will
    have been calculated, and `None` will be returned, indicating that the
    "onbuild" step should not be run.
    """
    return getattr(dist, "_versioningit_template_fields", None)



from __future__ import annotations
from collections.abc import Sequence
from datetime import datetime, timezone
import json
import os
from pathlib import Path
import re
import shlex
import subprocess
from typing import Any, Optional
from packaging.version import Version
from .errors import ConfigError, InvalidVersionError
from .logging import log


def str_guard(v: Any, fieldname: str) -> str:
    """
    If ``v`` is a `str`, return it; otherwise, raise a `ConfigError`.
    ``fieldname`` is an identifier for ``v`` to include in the error message.
    """
    if isinstance(v, str):
        return v
    else:
        raise ConfigError(f"{fieldname} must be set to a string")


def optional_str_guard(v: Any, fieldname: str) -> Optional[str]:
    """
    If ``v`` is a `str` or `None`, return it; otherwise, raise a `ConfigError`.
    ``fieldname`` is an identifier for ``v`` to include in the error message.
    """
    if v is None or isinstance(v, str):
        return v
    else:
        raise ConfigError(f"{fieldname} must be a string")


def list_str_guard(v: Any, fieldname: str) -> list[str]:
    """
    If ``v`` is a `list` of `str`\\s, return it; otherwise, raise a
    `ConfigError`.  ``fieldname`` is an identifier for ``v`` to include in the
    error message.
    """
    if isinstance(v, list) and all(isinstance(e, str) for e in v):
        return v
    else:
        raise ConfigError(f"{fieldname} must be a list of strings")


def bool_guard(v: Any, fieldname: str) -> bool:
    """
    If ``v`` is a `bool`, return it; otherwise, raise a `ConfigError`.
    ``fieldname`` is an identifier for ``v`` to include in the error message.
    """
    if isinstance(v, bool):
        return v
    else:
        raise ConfigError(f"{fieldname} must be set to a boolean")


def runcmd(*args: str | Path, **kwargs: Any) -> subprocess.CompletedProcess:
    """Run and log a given command"""
    arglist = [str(a) for a in args]
    log.debug("Running: %s", showcmd(arglist))
    kwargs.setdefault("check", True)
    return subprocess.run(arglist, **kwargs)


def readcmd(*args: str | Path, **kwargs: Any) -> str:
    """Run a command, capturing & returning its stdout"""
    s = runcmd(*args, stdout=subprocess.PIPE, text=True, **kwargs).stdout
    assert isinstance(s, str)
    return s.strip()


def get_build_date() -> datetime:
    """
    Return the current date & time as an aware UTC `~datetime.datetime`.  If
    :envvar:`SOURCE_DATE_EPOCH` is set, use that value instead (See
    <https://reproducible-builds.org/specs/source-date-epoch/>).
    """
    try:
        source_date_epoch = int(os.environ["SOURCE_DATE_EPOCH"])
    except (KeyError, ValueError):
        return datetime.now(timezone.utc)
    else:
        return fromtimestamp(source_date_epoch)


def fromtimestamp(ts: int) -> datetime:
    """
    Convert an integer number of seconds since the epoch to an aware UTC
    `~datetime.datetime`
    """
    return datetime.fromtimestamp(ts, tz=timezone.utc)


def strip_prefix(s: str, prefix: str) -> str:
    """
    If ``s`` starts with ``prefix``, return the rest of ``s`` after ``prefix``;
    otherwise, return ``s`` unchanged.
    """
    # cf. str.removeprefix, introduced in Python 3.9
    n = len(prefix)
    return s[n:] if s[:n] == prefix else s


def strip_suffix(s: str, suffix: str) -> str:
    """
    If ``s`` ends with ``suffix``, return the rest of ``s`` before ``suffix``;
    otherwise, return ``s`` unchanged.
    """
    # cf. str.removesuffix, introduced in Python 3.9
    n = len(suffix)
    return s[:-n] if s[-n:] == suffix else s


def parse_version_from_metadata(metadata: str) -> str:
    """
    Given a string containing Python packaging metadata, return the value of
    the :mailheader:`Version` field

    :raises ValueError: if there is no :mailheader:`Version` field
    """
    for line in metadata.splitlines():
        m = re.match(r"Version\s*:\s*", line)
        if m:
            return line[m.end() :].strip()
        elif not line:
            break
    raise ValueError("Metadata does not contain a Version field")


def showcmd(args: list) -> str:
    """
    Stringify the elements of ``args``, shell-quote them, and join the results
    with a space
    """
    return " ".join(shlex.quote(os.fsdecode(a)) for a in args)


def is_sdist(project_dir: str | Path) -> bool:
    """
    Performs a simplistic check whether ``project_dir`` (which presumably is
    not under version control) is an unpacked sdist by testing whether
    :file:`PKG-INFO` exists
    """
    if Path(project_dir, "PKG-INFO").exists():
        log.info(
            "Directory is not under version control, and PKG-INFO is present;"
            " assuming this is an sdist"
        )
        return True
    else:
        return False


def split_version(
    v: str, split_on: Optional[str] = None, double_quote: bool = True
) -> str:
    if split_on is None:
        split_on = r"[-_.+!]"
    parts = [int(p) if p.isdigit() else p for p in re.split(split_on, v) if p]
    return repr_tuple(parts, double_quote)


def split_pep440_version(
    v: str, double_quote: bool = True, epoch: Optional[bool] = None
) -> str:
    try:
        vobj = Version(v)
    except ValueError:
        raise InvalidVersionError(f"{v!r} is not a valid PEP 440 version")
    parts: list[str | int] = []
    if epoch or (vobj.epoch and epoch is None):
        parts.append(vobj.epoch)
    parts.extend(vobj.release)
    if vobj.pre is not None:
        phase, num = vobj.pre
        parts.append(f"{phase}{num}")
    if vobj.post is not None:
        parts.append(f"post{vobj.post}")
    if vobj.dev is not None:
        parts.append(f"dev{vobj.dev}")
    if vobj.local is not None:
        parts.append(f"+{vobj.local}")
    return repr_tuple(parts, double_quote)


def repr_tuple(parts: Sequence[str | int], double_quote: bool = True) -> str:
    strparts: list[str] = []
    for p in parts:
        if isinstance(p, int):
            strparts.append(str(p))
        elif double_quote:
            strparts.append(qqrepr(p))
        else:
            strparts.append(repr(p))
    return "(" + ", ".join(strparts) + ")"


def qqrepr(s: str) -> str:
    """Produce a repr(string) enclosed in double quotes"""
    return json.dumps(s, ensure_ascii=False)


def ensure_terminated(s: str) -> str:
    """Append a newline to ``s`` if it doesn't already end with one"""
    lines = s.splitlines(keepends=True)
    if not lines or lines[-1].splitlines() == [lines[-1]]:
        return s + "\n"
    else:
        return s



from __future__ import annotations
from dataclasses import dataclass
import os
from pathlib import Path
import subprocess
from typing import Any
from .core import VCSDescription
from .errors import NoTagError, NotVCSError
from .logging import log, warn_extra_fields
from .util import get_build_date, is_sdist, optional_str_guard, readcmd, runcmd


@dataclass
class HGRepo:
    """Methods for querying a Mercurial repository"""

    #: The repository's working tree or a subdirectory thereof
    path: str | Path

    def ensure_is_repo(self) -> None:
        """
        Test whether `path` is under Mercurial revision control; if it is not
        (or if Mercurial is not installed), raise a `NotVCSError`
        """
        try:
            runcmd(
                "hg",
                "--cwd",
                self.path,
                "files",
                ".",
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                env={**os.environ, "HGPLAIN": "1"},
            )
        except FileNotFoundError:
            raise NotVCSError(
                "hg not installed; assuming this isn't a Mercurial repository"
            )
        except subprocess.CalledProcessError:
            raise NotVCSError(f"{self.path} is not tracked by Mercurial")

    def read(self, *args: str, **kwargs: Any) -> str:
        """
        Run a Mercurial command with the given arguments in `path` and return
        the stripped stdout
        """
        return readcmd(
            "hg",
            "--cwd",
            self.path,
            *args,
            env={**os.environ, "HGPLAIN": "1"},
            **kwargs,
        )


def describe_hg(*, project_dir: str | Path, params: dict[str, Any]) -> VCSDescription:
    """Implements the ``"hg"`` ``vcs`` method"""
    params = params.copy()
    pattern = optional_str_guard(
        params.pop("pattern", None), "tool.versioningit.vcs.pattern"
    )
    default_tag = optional_str_guard(
        params.pop("default-tag", None), "tool.versioningit.vcs.default-tag"
    )
    warn_extra_fields(params, "tool.versioningit.vcs", ["pattern", "default-tag"])
    build_date = get_build_date()
    repo = HGRepo(project_dir)
    try:
        repo.ensure_is_repo()
    except NotVCSError:
        if not is_sdist(project_dir) and Path(project_dir, ".hg_archival.txt").exists():
            log.info("%s is a Mercurial archive; parsing .hg_archival.txt", project_dir)
            data = parse_hg_archival(Path(project_dir, ".hg_archival.txt"))
            if "tag" in data:
                tag = data["tag"]
                distance = 0
            else:
                tag = data["latesttag"]
                distance = int(data["changessincelatesttag"])
            branch = data["branch"]
            revision = data["node"]
            rev = revision[:12]
        else:
            raise
    else:
        # Use "{changes}" instead of "{distance}", as the former counts all
        # different commits across all parent paths (which is what `git
        # describe` does), while the latter is just the length of the longest
        # path.
        if pattern is None:
            template = "{latesttag() % '{tag}:{changes}:{node}\n'}"
        else:
            template = "{latesttag(" + repr(pattern) + ") % '{tag}:{changes}:{node}\n'}"
        tag, sdistance, revision = (
            repo.read("log", "-r", ".", "--template", template)
            .splitlines()[0]
            .split(":")
        )
        distance = int(sdistance)
        rev, _, branch = repo.read("id", "-i", "-b").partition(" ")
    if rev.endswith("+"):
        dirty = True
        rev = rev[:-1]
    else:
        dirty = False
    if tag == "null":
        if default_tag is not None:
            log.info("No latest tag; falling back to default tag %r", default_tag)
            tag = default_tag
            # Act as though the first commit is the one with the default tag,
            # i.e., don't count it (unless there is no first commit, of course)
            if distance > 0:
                distance -= 1
        else:
            raise NoTagError("No latest tag in Mercurial repository")
    if distance and dirty:
        state = "distance-dirty"
    elif distance:
        state = "distance"
    elif dirty:
        state = "dirty"
    else:
        state = "exact"
    return VCSDescription(
        tag=tag,
        state=state,
        branch=branch,
        fields={
            "distance": distance,
            "rev": rev,
            "revision": revision,
            "build_date": build_date,
            "vcs": "h",
            "vcs_name": "hg",
        },
    )


def parse_hg_archival(path: Path) -> dict[str, str]:
    data: dict[str, str] = {}
    with path.open() as fp:
        for line in fp:
            key, _, value = line.strip().partition(": ")
            data.setdefault(key, value)
    return data



from __future__ import annotations
import argparse
import logging
import os
import subprocess
import sys
import traceback
from typing import Optional
from . import __version__
from .core import get_next_version, get_version
from .errors import Error
from .logging import get_env_loglevel, log
from .util import showcmd


def main(argv: Optional[list[str]] = None) -> None:
    parser = argparse.ArgumentParser(
        description="Show the version of a versioningit-enabled project"
    )
    parser.add_argument(
        "-n",
        "--next-version",
        action="store_true",
        help="Show the next version after the current VCS tag",
    )
    parser.add_argument(
        "--traceback", action="store_true", help="Show full traceback on library error"
    )
    parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Show more log messages"
    )
    parser.add_argument(
        "-w", "--write", action="store_true", help="Write version to configured file"
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument("project_dir", nargs="?", default=os.curdir)
    args = parser.parse_args(argv)
    env_loglevel = get_env_loglevel()
    if args.verbose == 0:
        if env_loglevel is None:
            log_level = logging.WARNING
        else:
            log_level = env_loglevel
    else:
        if args.verbose == 1:
            log_level = logging.INFO
        else:
            log_level = logging.DEBUG
        if env_loglevel is not None:
            log_level = min(log_level, env_loglevel)
    logging.basicConfig(
        format="[%(levelname)-8s] %(name)s: %(message)s",
        level=log_level,
    )
    try:
        if args.next_version:
            print(get_next_version(args.project_dir))
        else:
            print(get_version(args.project_dir, write=args.write, fallback=True))
    except Error as e:
        if args.traceback:
            traceback.print_exc()
        else:
            print(f"versioningit: {type(e).__name__}: {e}", file=sys.stderr)
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        if args.traceback:
            traceback.print_exc()
        else:
            if isinstance(e.cmd, list):
                cmd = showcmd(e.cmd)
            else:
                cmd = os.fsdecode(e.cmd)
            log.error("%s: command returned %d", cmd, e.returncode)
        sys.exit(e.returncode)


if __name__ == "__main__":
    main()  # pragma: no cover



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional
from .core import get_template_fields_from_distribution, run_onbuild
from .logging import init_logging, log

if TYPE_CHECKING:
    from setuptools import Command


def get_cmdclasses(
    bases: Optional[dict[str, type[Command]]] = None
) -> dict[str, type[Command]]:
    """
    .. versionadded:: 1.1.0

    Return a `dict` of custom setuptools `Command` classes, suitable for
    passing to the ``cmdclass`` argument of `setuptools.setup()`, that run the
    ``onbuild`` step for the project when building an sdist or wheel.
    Specifically, the `dict` contains a subclass of
    `setuptools.command.sdist.sdist` at the ``"sdist"`` key and a subclass of
    `setuptools.command.build_py.build_py` at the ``"build_py"`` key.

    A `dict` of alternative base classes can optionally be supplied; if the
    `dict` contains an ``"sdist"`` entry, that entry will be used as the base
    class for the customized ``sdist`` command, and likewise for
    ``"build_py"``.  All other classes in the input `dict` are passed through
    unchanged.
    """
    # Import setuptools here so there isn't a slowdown from importing it
    # unconditionally whenever versioningit is imported
    from setuptools.command.build_py import build_py
    from setuptools.command.sdist import sdist

    cmds = {} if bases is None else bases.copy()

    sdist_base = cmds.get("sdist", sdist)

    class VersioningitSdist(sdist_base):  # type: ignore[valid-type,misc]
        def make_release_tree(self, base_dir: str, files: Any) -> None:
            super().make_release_tree(base_dir, files)
            init_logging()
            template_fields = get_template_fields_from_distribution(self.distribution)
            if template_fields is not None:
                PROJECT_ROOT = Path().resolve()
                log.debug("Running onbuild step; cwd=%s", PROJECT_ROOT)
                run_onbuild(
                    project_dir=PROJECT_ROOT,
                    build_dir=base_dir,
                    is_source=True,
                    template_fields=template_fields,
                )
            else:
                log.debug(
                    "Appear to be building from an sdist; not running onbuild step"
                )

    cmds["sdist"] = VersioningitSdist

    build_py_base = cmds.get("build_py", build_py)

    class VersioningitBuildPy(build_py_base):  # type: ignore[valid-type,misc]
        editable_mode: bool = False

        def run(self) -> None:
            super().run()
            init_logging()
            template_fields = get_template_fields_from_distribution(self.distribution)
            if not self.editable_mode and template_fields is not None:
                PROJECT_ROOT = Path().resolve()
                log.debug("Running onbuild step; cwd=%s", PROJECT_ROOT)
                run_onbuild(
                    project_dir=PROJECT_ROOT,
                    build_dir=self.build_lib,
                    is_source=False,
                    template_fields=template_fields,
                )
            else:
                log.debug(
                    "Appear to be building from an sdist; not running onbuild step"
                )

    cmds["build_py"] = VersioningitBuildPy

    return cmds



from __future__ import annotations
from pathlib import Path
import re
from typing import Any
from .errors import ConfigError
from .logging import log, warn_extra_fields
from .util import bool_guard, ensure_terminated, optional_str_guard, str_guard


def replace_version_onbuild(
    *,
    build_dir: str | Path,
    is_source: bool,
    template_fields: dict[str, Any],
    params: dict[str, Any],
) -> None:
    """Implements the ``"replace-version"`` ``onbuild`` method"""

    DEFAULT_REGEX = r"^\s*__version__\s*=\s*(?P<version>.*)"
    DEFAULT_REPLACEMENT = '"{version}"'

    params = params.copy()
    source_file = str_guard(
        params.pop("source-file", None), "tool.versioningit.onbuild.source-file"
    )
    build_file = str_guard(
        params.pop("build-file", None), "tool.versioningit.onbuild.build-file"
    )
    encoding = str_guard(
        params.pop("encoding", "utf-8"), "tool.versioningit.onbuild.encoding"
    )
    regex = str_guard(
        params.pop("regex", DEFAULT_REGEX), "tool.versioningit.onbuild.regex"
    )
    try:
        rgx = re.compile(regex)
    except re.error as e:
        raise ConfigError(f"tool.versioningit.onbuild.regex: Invalid regex: {e}")
    require_match = bool_guard(
        params.pop("require-match", False), "tool.versioningit.onbuild.require-match"
    )
    replacement = str_guard(
        params.pop("replacement", DEFAULT_REPLACEMENT),
        "tool.versioningit.onbuild.replacement",
    )
    append_line = optional_str_guard(
        params.pop("append-line", None), "tool.versioningit.onbuild.append-line"
    )
    warn_extra_fields(
        params,
        "tool.versioningit.onbuild",
        [
            "source-file",
            "build-file",
            "encoding",
            "regex",
            "require-match",
            "replacement",
            "append-line",
        ],
    )

    path = Path(build_dir, source_file if is_source else build_file)
    log.info("Updating version in file %s", path)
    lines = path.read_text(encoding=encoding).splitlines(keepends=True)
    for i, ln in enumerate(lines):
        m = rgx.search(ln)
        if m:
            log.debug("onbuild.regex matched file on line %d", i + 1)
            vgroup: str | int
            if "version" in m.groupdict():
                vgroup = "version"
            else:
                vgroup = 0
            if m[vgroup] is None:
                raise RuntimeError(
                    "'version' group in tool.versioningit.onbuild.regex did"
                    " not participate in match"
                )
            newline = ensure_terminated(
                ln[: m.start(vgroup)]
                + m.expand(replacement.format_map(template_fields))
                + ln[m.end(vgroup) :]
            )
            log.debug("Replacing line %r with %r", ln, newline)
            lines[i] = newline
            break
    else:
        if require_match:
            raise RuntimeError(f"onbuild.regex did not match any lines in {path}")
        elif append_line is not None:
            log.info(
                "onbuild.regex did not match any lines in the file; appending line"
            )
            if lines:
                lines[-1] = ensure_terminated(lines[-1])
            lines.append(ensure_terminated(append_line.format_map(template_fields)))
        else:
            log.info(
                "onbuild.regex did not match any lines in the file; leaving unmodified"
            )
            return
    path.unlink()  # In case of hard links
    path.write_text("".join(lines), encoding=encoding)
